'`matflow_defdap.__init__.py`'

from functools import partial

from matflow_defdap._version import __version__

from matflow.extensions import (
    input_mapper,
    output_mapper,
    cli_format_mapper,
    register_output_file,
    func_mapper,
    software_versions,
    sources_mapper,
)

SOFTWARE = 'DefDAP'

input_mapper = partial(input_mapper, software=SOFTWARE)
output_mapper = partial(output_mapper, software=SOFTWARE)
cli_format_mapper = partial(cli_format_mapper, software=SOFTWARE)
register_output_file = partial(register_output_file, software=SOFTWARE)
func_mapper = partial(func_mapper, software=SOFTWARE)
software_versions = partial(software_versions, software=SOFTWARE)
sources_mapper = partial(sources_mapper, software=SOFTWARE)

def main_func(func):
    '''Decorator used to annotate which function within a snippet is the main function.'''
    def main_inner(*args, **kwargs):
        func(*args, **kwargs)
    return main_inner

# Import any modules that contain functions decorated by the matflow function decorators.
# This import must come after the above `partial`s.
from matflow_defdap import main



import numpy as np
from defdap.quat import Quat
from scipy.stats import mode
from scipy.ndimage import zoom

from matflow_defdap import main_func


@main_func
def get_DIC_image(DicMap, scaling_factor):

    # Construct an array of Euler angles
    grain_quats = np.empty((len(DicMap), 4))

    # Transformation orientations from EBSD orientation reference frame
    # to EBSD spatial reference frame
    frame_transform = Quat.fromAxisAngle(np.array((1, 0, 0)), np.pi)

    if DicMap.ebsdMap.crystalSym == 'hexagonal':
        # Convert hex convention from y // a2 of EBSD map to x // a1 for DAMASK
        hex_transform = Quat.fromAxisAngle(np.array([0, 0, 1]), -np.pi/6)
        for i, grain in enumerate(DicMap):
            grain_quats[i] = (hex_transform * grain.ebsdGrain.refOri * frame_transform).quatCoef

    else:
        for i, grain in enumerate(DicMap):
            grain_quats[i] = (grain.ebsdGrain.refOri * frame_transform).quatCoef

    # Filter out -1 (grain boundary points) and -2 (too small grains)
    # values in the grain image
    grain_image = DicMap.grains
    remove_boundary_points(grain_image)
    remove_small_grain_points(grain_image)
    remove_boundary_points(grain_image)
    remove_boundary_points(grain_image, force_remove=True)

    # scale down image if needed
    if scaling_factor != 1:
        grain_image = zoom(grain_image, scaling_factor, order=0, 
                           prefilter=False, mode='nearest')

    # downstream expects grain numbering to start at 0 not 1
    grain_image -= 1

    try:
        dic_scale = DicMap.scale
    except ValueError:
        dic_scale = None
    DIC_image = {
        'orientations': {
            'type': 'quat',
            'unit_cell_alignment': {'x': 'a'},
            'quaternions': grain_quats,
            'P': 1,  # DefDAP uses P=+1 (e.g see `defdap.quat.Quat.__mul__`)
            'quat_component_ordering': 'scalar-vector',
        },
        'grains': grain_image,
        'scale': dic_scale,
    }

    return DIC_image


def select_area(i, j, grain_image):
    i_min, i_max = 1, 1
    j_min, j_max = 1, 1

    on_edge = 0

    if i == 0:
        i_min = 0
        on_edge += 1
    elif i == grain_image.shape[0]-1:
        i_max = 0
        on_edge += 1

    if j == 0:
        j_min = 0
        on_edge += 1
    elif j == grain_image.shape[1]-1:
        j_max = 0
        on_edge += 1

    # select 3x3 region around point
    area = grain_image[i-i_min:i+i_max+1, j-j_min:j+j_max+1]

    return area, on_edge


def remove_boundary_points(grain_image, force_remove=False, 
                           max_iterations=200):
    num_bad_prev = 0
    iteration = 0
    while True:
        num_bad = np.count_nonzero(grain_image == -1)
        if num_bad == 0:
            # No bad values left, done
            print("All bad points removed.")
            break
        elif num_bad == num_bad_prev:
            # Not removing any more
            print("Number of bad points is not decreasing!")
            break
        if iteration > max_iterations:
            print("Max iterations.")
            break

        iteration += 1
        print("Starting iteration {}, num bad: {}".format(iteration, num_bad))

        grain_image_new = np.copy(grain_image)

        # because of how boundaries are defined, best to take from -ve side
        for i, j in zip(*np.where(grain_image == -1)):
            if i != 0 and j != 0 and grain_image[i-1, j] > 0 and grain_image[i, j-1] > 0:
                # if both left and above pixels defined, check the diagonal
                if grain_image[i-1, j-1] == grain_image[i-1, j]:
                    grain_image_new[i, j] = grain_image[i-1, j]

                elif grain_image[i-1, j-1] == grain_image[i, j-1]: 
                    grain_image_new[i, j] = grain_image[i, j-1]

                else:
                    # give up, try in next iteration

                    if force_remove:
                        area, on_edge = select_area(i, j, grain_image)
                        area = area.flatten()
                        area = area[np.where(area > 0)]   # remove -1 and -2

                        mode_vals, mode_counts = mode(area)
                        for mode_val, mode_count in zip(mode_vals, mode_counts):
                #         mode_val, mode_count = mode_vals[0], mode_counts[0]
    #                         if mode_count >= num_neighbours:
                            grain_image_new[i, j] = mode_val
                            break


            elif i != 0 and grain_image[i-1, j] > 0:
                grain_image_new[i, j] = grain_image[i-1, j]

            elif j != 0 and grain_image[i, j-1] > 0:
                grain_image_new[i, j] = grain_image[i, j-1]

            # give up, try in next iteration

        num_bad_prev = num_bad
        # [:, :] required to update the array passed in
        grain_image[:, :] = grain_image_new


def remove_small_grain_points(grain_image, max_iterations=200):
    # num_neighbours - must have at least this many pixels surrounding
    # start checking for 8 neighbours, then 7 until 2
    all_done = False
    for num_neighbours in list(range(8, 1, -1)):
        print(f"Starting iterations with at least {num_neighbours} equal neighbours")

        num_bad_prev = 0
        iteration = 0
        while True:
            num_bad = np.count_nonzero(grain_image == -2)
            if num_bad == 0:
                # No bad values left, done
                print("All bad points removed.")
                all_done = True
                break
            elif num_bad == num_bad_prev:
                # Not removing any more
                print("Number of bad points is not decreasing!")
                break
            if iteration > max_iterations:
                print("Max iterations.")
                break

            iteration += 1
            print("Starting iteration {}, num bad: {}".format(iteration, num_bad))

            grain_image_new = np.copy(grain_image)

            for i, j in zip(*np.where(grain_image == -2)):

                area, on_edge = select_area(i, j, grain_image)
                area = area.flatten()
                area = area[np.where(area > 0)]   # remove -1 and -2

                mode_vals, mode_counts = mode(area)
                for mode_val, mode_count in zip(mode_vals, mode_counts):
                    if mode_count >= num_neighbours:
                        grain_image_new[i, j] = mode_val
                        break

            num_bad_prev = num_bad
            # [:, :] required to update the array passed in
            grain_image[:, :] = grain_image_new

        if all_done:
            break



import defdap.hrdic as hrdic

from matflow_defdap import main_func


@main_func
def load_DIC_map(root_path, dic_filename, dic_crop, dic_scale):
    'Load in DIC, crop and set scale.'

    DicMap = hrdic.Map(root_path, dic_filename)
    if dic_crop is not None:
        DicMap.setCrop(
            xMin=dic_crop[0],
            xMax=dic_crop[1],
            yMin=dic_crop[2],
            yMax=dic_crop[3]
        )
    if dic_scale is not None:
        DicMap.setScale(dic_scale)

    return DicMap



from matflow_defdap import main_func


@main_func
def link_EBSD_DIC_maps(EbsdMap, DicMap, dic_homog_points, dic_min_grain_size,
                       ebsd_homog_points, transform_type):

    DicMap.homogPoints = dic_homog_points
    EbsdMap.homogPoints = ebsd_homog_points
    DicMap.linkEbsdMap(EbsdMap, transformType=transform_type)

    DicMap.findGrains(minGrainSize=dic_min_grain_size)



from pathlib import Path

import defdap.ebsd as ebsd

from matflow_defdap import main_func


@main_func
def load_EBSD_map(root_path, ebsd_filename, ebsd_flip_vert,
                  ebsd_boundary_tol, ebsd_min_grain_size):
    'Load EBSD map and detect grains.'

    EbsdMap = ebsd.Map(Path(root_path).joinpath(ebsd_filename))

    # Flip EBSD map in vertical direction if needed
    if ebsd_flip_vert:
        EbsdMap.eulerAngleArray = EbsdMap.eulerAngleArray[:, ::-1, ::-1]
        EbsdMap.bandContrastArray = EbsdMap.bandContrastArray[::-1, ::-1]
        EbsdMap.phaseArray = EbsdMap.phaseArray[::-1, ::-1]

    EbsdMap.buildQuatArray()

    EbsdMap.findBoundaries(boundDef=ebsd_boundary_tol)
    EbsdMap.findGrains(minGrainSize=ebsd_min_grain_size)
    EbsdMap.calcGrainAvOris()

    return EbsdMap



'`matflow_defdap.main.py`'

import hickle

from matflow_defdap import (
    input_mapper,
    output_mapper,
    cli_format_mapper,
    register_output_file,
    func_mapper,
    software_versions,
    sources_mapper
)
from matflow_defdap.scripting import get_wrapper_script


@sources_mapper(task='load_microstructure', method='EBSD+DIC',
                script='load_microstructure')
def load_microstructure_EBSD_DIC():

    script_name = 'load_microstructure.py'
    snippets = [
        {'name': 'load_EBSD_map.py'},
        {'name': 'load_DIC_map.py'},
        {'name': 'link_EBSD_DIC_maps.py'},
        {'name': 'get_DIC_image.py'},
    ]
    outputs = ['DIC_image']
    out = {
        'script': {
            'content': get_wrapper_script(script_name, snippets, outputs),
            'filename': script_name,
        }
    }
    return out


@input_mapper(input_file='inputs.hdf5', task='load_microstructure', method='EBSD+DIC')
def write_param_file(path, DIC, EBSD, transform_type, root_path, scaling_factor):
    obj = {
        'dic_filename': DIC['filename'],
        'dic_crop': DIC.get('crop', None),
        'dic_scale': DIC.get('scale', None),
        'dic_min_grain_size': DIC.get('min_grain_size', 10),
        'dic_homog_points': DIC['homog_points'],
        'ebsd_filename': EBSD['filename'],
        'ebsd_flip_vert': EBSD.get('flip_vert', False),
        'ebsd_boundary_tol': EBSD.get('boundary_tol', 10),
        'ebsd_min_grain_size': EBSD.get('min_grain_size', 10),
        'ebsd_homog_points': EBSD['homog_points'],
        'transform_type': transform_type,
        'root_path': root_path,
        'scaling_factor': scaling_factor,
    }
    hickle.dump(obj, path)


@output_mapper(
    output_name='microstructure_image',
    task='load_microstructure',
    method='EBSD+DIC',
)
def read_DIC_image_file(path):
    return hickle.load(path)


@sources_mapper(task='load_microstructure', method='EBSD',
                script='load_microstructure')
def load_microstructure_EBSD():

    script_name = 'load_microstructure.py'
    snippets = [
        {'name': 'load_EBSD_map.py'},
        {'name': 'get_EBSD_image.py'},
    ]
    outputs = ['EBSD_image']
    out = {
        'script': {
            'content': get_wrapper_script(script_name, snippets, outputs),
            'filename': script_name,
        }
    }
    return out


@input_mapper(input_file='inputs.hdf5', task='load_microstructure', method='EBSD')
def write_param_file_2(path, EBSD, root_path, scaling_factor):
    obj = {
        'ebsd_filename': EBSD['filename'],
        'ebsd_flip_vert': EBSD.get('flip_vert', False),
        'ebsd_boundary_tol': EBSD.get('boundary_tol', 10),
        'ebsd_min_grain_size': EBSD.get('min_grain_size', 10),
        'root_path': root_path,
        'scaling_factor': scaling_factor,
    }
    hickle.dump(obj, path)


@output_mapper(
    output_name='microstructure_image',
    task='load_microstructure',
    method='EBSD',
)
def read_EBSD_image_file(path):
    return hickle.load(path)



import re
from textwrap import dedent

import black
import autopep8

from pkg_resources import resource_string


def get_snippet(name, decorator=True):
    'Get a Python snippet function (as a string) from the snippets directory.'
    out = resource_string('matflow_defdap', f'snippets/{name}').decode()
    if not decorator:
        # Remove the `@main_func` decorator and import.
        remove_lns = ['from matflow_defdap import main_func', '@main_func']
        for i in remove_lns:
            out = ''.join(out.split(i))
    return out


def parse_python_func_return(func_str):
    """Get a list of the variable names in a Python function return statement.

    The return statement may return a tuple (with parenthesis or not) or a single variable.

    """

    out = []
    match = re.search(r'return \(*([\S\s][^\)]+)\)*', func_str)
    if match:
        match_clean = match.group(1).strip().strip(',')
        out = [i.strip() for i in match_clean.split(',')]

    return out


def parse_python_func_imports(func_str):
    """Get a list of import statement lines from a (string) Python function."""

    import_lines = func_str.split('def ')[0].strip()
    match = re.search(r'((?:import|from)[\S\s]*)', import_lines)
    out = []
    if match:
        out = match.group(1).splitlines()

    return out


def extract_snippet_main(snippet_str):
    """Extract only the snippet main function (plus imports), as annotated by the
    `@mainfunc` decorator."""

    func_start_pat = r'((?:@main_func\n)?def\s(?:.*)\((?:[\s\S]*?)\):)'

    func_split_snip = re.split(func_start_pat, snippet_str)
    imports = func_split_snip[0]
    main_func_dec_str = '@main_func'

    main_func_str = None
    for idx in range(1, len(func_split_snip[1:]), 2):
        func_str = func_split_snip[idx] + func_split_snip[idx + 1]
        if main_func_dec_str in func_str:
            if main_func_str:
                msg = (f'`{main_func_dec_str}` should decorate only one function within '
                       f'the snippet.')
                raise ValueError(msg)
            else:
                main_func_str = func_str.lstrip(f'{main_func_dec_str}\n')

    imports = ''.join(imports.split('from matflow_defdap import main_func'))

    return imports + '\n' + main_func_str


def get_snippet_signature(script_name):
    'Get imports, inputs and outputs of a Python snippet function.'

    snippet_str = get_snippet(script_name)
    snippet_str = extract_snippet_main(snippet_str)

    def_line = re.search(r'def\s(.*)\(([\s\S]*?)\):', snippet_str).groups()
    func_name = def_line[0]
    func_ins = [i.strip() for i in def_line[1].split(',')]

    if script_name != func_name + '.py':
        msg = ('For simplicity, the snippet main function name should be the same as the '
               'snippet file name.')
        raise ValueError(msg)

    func_outs = parse_python_func_return(snippet_str)
    func_imports = parse_python_func_imports(snippet_str)

    out = {
        'name': func_name,
        'imports': func_imports,
        'inputs': func_ins,
        'outputs': func_outs,
    }
    return out


def get_snippet_call(script_name):
    sig = get_snippet_signature(script_name)
    outs_fmt = ', '.join(sig['outputs'])
    ins_fmt = ', '.join(sig['inputs'])
    ret = f'{sig["name"]}({ins_fmt})'
    if outs_fmt:
        ret = f'{outs_fmt} = {ret}'
    return ret


def get_wrapper_script(script_name, snippets, outputs):

    ind = '    '
    sigs = [get_snippet_signature(i['name']) for i in snippets]
    all_ins = [j for i in sigs for j in i['inputs']]
    all_outs = [j for i in sigs for j in i['outputs']]

    for i in outputs:
        if i not in all_outs:
            raise ValueError(f'Cannot output "{i}". No functions return this name.')

    # Required inputs are those that are not output by any snippet
    req_ins = list(set(all_ins) - set(all_outs))
    req_ins_fmt = ', '.join(req_ins)

    main_sig = [f'def main({req_ins_fmt}):']
    main_body = [ind + get_snippet_call(i['name']) for i in snippets]
    main_outs = ['\n' + ind + f'return {", ".join([i for i in outputs])}']
    main_func = main_sig + main_body + main_outs

    req_imports = [
        'import sys',
        'import hickle',
        'from pathlib import Path',
    ]
    out = req_imports
    out += main_func
    snippet_funcs = '\n'.join([get_snippet(i['name'], decorator=False) for i in snippets])

    out = '\n'.join(out) + '\n' + snippet_funcs + '\n'
    out += dedent('''\
        if __name__ == '__main__':        
            inputs = hickle.load(sys.argv[1])
            outputs = main(**inputs)
            hickle.dump(outputs, 'outputs.hdf5')

    ''')

    out = autopep8.fix_code(out)
    out = black.format_str(out, mode=black.FileMode())

    return out
